#
# packing.rb
#
#   Copyright (c) 1999 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#

require 'ftools'

require 'amstd/common'


module Packing


  class Scope

    def initialize( name, stack )
      @name = name.dup
      @stack = stack
      @values = {}

      @verbose = false
    end

    attr :name

    def []( name )
      @values[ _name2str name ]
    end

    def []=( name, val )
      @values[ _name2str name ] = val
    end


    def set( name, val )
      if String === name and (a = name.split('.')).size > 1 then
        @stack[ a[0] ][ a[1] ] = val
      else
        self[ name ] = val
      end
    end

    alias s set

    def get( name )
      if String === name and (a = name.split('.')).size > 1 then
        ret = @stack[ a[0] ][ a[1] ]
        warn "#{name} is nil" unless ret
      else
        ret = self[ name ]
        warn "#{self.name}.#{_name2str(name)} is nil" unless ret
      end

      ret
    end

    alias g get

    def warn( mes )
      $stderr.print 'Warning: '
      $stderr.puts mes
    end


    def get_env( name )
      @stack[ name ]
    end

    alias e get_env


    def expand( fn )
      File.expand_path fn
    end


    def mkdir( dname )
      return if File.directory? dname
      Dir.mkdir dname
    end

    def newdir( dname )
      mkdir dname
      dname
    end

    def foreach_f( recv, sim, key, *args )
      if Array === key then
        key.each do |k|
          recv.send sim, k, *args
        end
      else
        recv.send sim, key, *args
      end
    end

    def install( from, to, mode = nil )
      File.install from, to, mode, @verbose
    end

    def cp( from, to )
      foreach_f File, :cp, from, to, @verbose
    end

    def mv( from, to )
      foreach_f File, :mv, from, to, @verbose
    end

    def rm( fn )
      foreach_f File, :unlink, fn
    end

    def rm_f( fn )
      foreach_f File, :rm_f, fn
    end

    def rm_rf( dn )
      begin
        if Array === dn then
          dn.each do |f|
            do_rmrf f if File.directory? f
          end
        else
          do_rmrf dn if File.directory? dn
        end
      rescue
        puts "error in rm_rf"
        $stderr.puts $!
      end
    end

    def do_rmrf( dn )
      foreach_fullpath( dn ) do |fn|
        if File.directory? fn then
          do_rmrf fn
        else
          File.rm_f fn
        end
      end
      Dir.rmdir dn
    end

    def foreach_fullpath( dn )
      Dir.foreach( dn ) do |fn|
        next if /\A\.\.?\z/o === fn
        yield dn + '/' + fn
      end
    end

    def chmod( mod, f )
      File.chmod mod, *f
    end


    def a( arg )     # to Absolute path
      d( (g :dir), arg )
    end

    def d( dir, arg )    # in directory
      if Array === arg then
        arg.collect{|f| dir + '/' + f }
      else
        "#{dir}/#{arg}"
      end
    end

  end


  class ScopeError < StandardError ; end

  class ScopeStack

    def initialize
      @vals = {}
      @current = nil
      @prev = nil
      @stack = []
    end


    attr :current

    def []( name )
      @vals[ _name2str( name ) ]
    end

    def enter( name )
      name = _name2str( name )
      if s = @vals[ name ] then
        @current = s
      else
        @vals[ name ] = @current = Scope.new( name, self )
      end
      @stack.push @current

      @current
    end

    def leave( name = nil )
      name = _name2str( name ) if name
      unless s = @stack.pop then
        raise ScopeError, "too many leave than enter"
      end
      if name and s.name != name then
        puts "warning: leave from #{s.name}, but #{s.name} is required"
      end
      @prev = @current
      @current = @stack[-1]

      s
    end

    def invoke( task )
      @prev.send task
    end

  end


  MAIN_STACK = ScopeStack.new

  def environ( envname, &block )
    cur = MAIN_STACK.enter( envname )
    cur.instance_eval &block
    MAIN_STACK.leave envname
  end

  def invoke( task )
    MAIN_STACK.invoke task
  end

end


include Packing
